/************************************************************************************
 * WrightEagle (Soccer Simulation League 2D)                                        *
 * BASE SOURCE CODE RELEASE 2009                                                    *
 * Copyright (C) 1998-2009 WrightEagle 2D Soccer Simulation Team,                   *
 *                         Multi-Agent Systems Lab.,                                *
 *                         School of Computer Science and Technology,               *
 *                         University of Science and Technology of China, China.    *
 *                                                                                  *
 * This program is free software; you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License                      *
 * as published by the Free Software Foundation; either version 2                   *
 * of the License, or (at your option) any later version.                           *
 *                                                                                  *
 * This program is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    *
 * GNU General Public License for more details.                                     *
 *                                                                                  *
 * You should have received a copy of the GNU General Public License                *
 * along with this program; if not, write to the Free Software                      *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.  *
 ************************************************************************************/

#ifndef __Coach_H__
#define __Coach_H__

#include <set>
#include "Client.h"
#include "Types.h"
#include "TeamGraphic.h"

class Coach: public Client
{
public:
	/**
	* 캯
	*/
	Coach();
	~Coach();

	/*
	 * The main loop for coach.
	 */
	void Run();

	/*
	 * Funciion for coach to send options to server such as player types
	 * and other initial commands.
	 */
	void SendOptionToServer();

	/*
	 * This class is for coach to choose player types.
	 */
	class PlayerTypePriority{
	public:
		Unum num;
		PlayerRole role[FT_Max];
		int priority;
		bool operator <(PlayerTypePriority& other){return priority < other.priority;}
	};
private:

	/*
	 * Funciton for the online coach to decide what to say to direct the match.
	 */
	void DoDecisionMaking();

	/*
	 * Judge if communication is possible by the rule
	 */
	bool ShouldCommunicate();

	int  mTeammatePlayerType[TEAMSIZE + 1];

	bool mIsPlayerTypesDone;
	bool mIsCycleInfoSent;
	int  mTimeInterval;

	/*
	 * Send the team graphic to server.
	 */
    void SendTeamGraphic();

    /*
     * Create the team graphic by using a certain .xpm file point by point.
     * \param x the x position of the pixel point.
     * \param y the y position of the pixel point.
     * \param team_graphic the graphic to create.
     * \return if the graphic successfully created.
     */
    bool DoTeamGraphic(const int x, const int y, const TeamGraphic & team_graphic);

    TeamGraphic mTeamGraphic;
};

#endif
