/************************************************************************************
 * WrightEagle (Soccer Simulation League 2D)                                        *
 * BASE SOURCE CODE RELEASE 2009                                                    *
 * Copyright (C) 1998-2009 WrightEagle 2D Soccer Simulation Team,                   *
 *                         Multi-Agent Systems Lab.,                                *
 *                         School of Computer Science and Technology,               *
 *                         University of Science and Technology of China, China.    *
 *                                                                                  *
 * This program is free software; you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License                      *
 * as published by the Free Software Foundation; either version 2                   *
 * of the License, or (at your option) any later version.                           *
 *                                                                                  *
 * This program is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    *
 * GNU General Public License for more details.                                     *
 *                                                                                  *
 * You should have received a copy of the GNU General Public License                *
 * along with this program; if not, write to the Free Software                      *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.  *
 ************************************************************************************/

#ifndef __DynamicDebug_H__
#define __DynamicDebug_H__


#include <stdio.h>
#include <iostream>
#include <fstream>
#include <vector>
#include "Observer.h"
#include "PlayerParam.h"

class DynamicDebug
{
	DynamicDebug();

	struct MessageFileHead
	{
		char    mHeadFlag[2]; // ʶǷǶ̬Լ¼Ϣļ
		Time    mMaxCycle; // ļм¼
		long    mIndexTableSize; // С
		long    mIndexTableOffset; // λ
		long    mParserTableSize;
		long    mParserTableOffset;
		long    mDecisionTableSize;
		long    mDecisionTableOffset;
		long    mCommandSendTableSize;
		long    mCommandSendTableOffset;

		MessageFileHead(): mMaxCycle( -10 )
		{

		}
	};

	struct MessageIndexTableUnit
	{
		Time    mServerTime; // ݶӦ
		int     mDataSize; // ݳ
		long    mDataOffset; // ݵĴ洢λ
		long    mTimeOffset; // ʱĴ洢λ
	};

	struct Message
	{
		MessageType mType;
		std::string mString;

		Message(MessageType type, const char *msg): mType(type), mString(msg)
		{

		}
	};

public:
    /**
     * 캯
     */
    ~DynamicDebug();

    /**
     * ʵʼ
     */
    static DynamicDebug & instance();

    /**
     * Initial the observer.
     * @param pObserver the point to the Observer.
     */
    void Initial(Observer *pObserver);

    /**
     * 溯Ҫ¼serverϢʱõĽӿ
     * AddMessage()һҪŵ̵߳ںǰã̬Եõϵͳʱʱ
     * 磬ҪParser::Parse()Client::Decision()CommandSend::Run()֮ǰ
     */
    
    // Funcitons below are interfaces to be used to log the server message.
    // The function AddMessage() must be used before the entrance functions 
    // such as Parser::Parse(), Client::Decision(), CommandSend::Run() of each
    // threads, or the system time will get errors when dynamic debugging.
    /**
     * Log the server message.
     * @param msg the server message logged.
     * @param msg_type the type of message to be added.
     */
    void AddMessage(const char *msg, MessageType msg_type);

    /**
     * Add the time parser.
     * @param time the time.
     */
    void AddTimeParser(timeval &time);

    /**
     * Add the time when the decision is done.
     * @param time the time evaluation.
     */
    void AddTimeDecision(timeval &time);

    /**
     * Add the time when the command is sent.
     * @param time the time evaluation.
     */
    void AddTimeCommandSend(timeval &time);

    /**
     * 溯Ƕ̬ʱõĽӿ
     */
    // The following functions are interfaces for dynamic debugging.
    /**
     * Run with a certain piece of message.
     * @param msg the message to run.
     * @return the message type.
     */
    MessageType Run(char *msg);

    /**
     * Get the Message.
     * @param msg the message to get.
     * @return the message type.
     */
    MessageType GetMessage(char *msg);

    /**
     * Find a certain cycle.
     * @param cycle the cycle to find.
     * @return true if the certain cycle found.
     */
    bool FindCycle(int cycle);

    /**
     * Get the time of parser.
     * @return the time of parser.
     */
    timeval GetTimeParser();

    /**
     * Get the time of decision.
     * @return the time of decision.
     */
    timeval GetTimeDecision();

    /**
     * Get the time of sending the command.
     * @return the time of sending the command.
     */
    timeval GetTimeCommandSend();

private:
    /**
     * Save the server message.
     */
    void Flush();

private:
    Observer    *mpObserver; // WorldModelָ
    bool        mInitialOK; // ǷѾʼ

    // ļͷϢ¼4Ϣ
    MessageFileHead mFileHead;

    // 4¼ͱ4Ϣ
    std::vector<MessageIndexTableUnit>  mIndexTable;
    std::vector<timeval>                mParserTimeTable;
    std::vector<timeval>                mDecisionTimeTable;
    std::vector<timeval>                mCommandSendTimeTable;

    std::vector<Message>                mMessageTable;

    // 4ָڶдļʱʹ
    MessageIndexTableUnit   *mpIndex;
    timeval                 *mpParserTime;
    timeval                 *mpDecisionTime;
    timeval                 *mpCommandSendTime;

    // ǰȡĵԪ
    MessageIndexTableUnit   *mpCurrentIndex;

    // ļ
    ThreadMutex     mFileMutex;
    FILE            *mpFile;
    std::ifstream   *mpFileStream;
    std::streambuf  *mpStreamBuffer;

    bool mRunning; // Ƿ
    bool mShowMessage; // ǷϢ
    Time mRuntoCycle; // ҪRun
};

#endif

