/************************************************************************************
 * WrightEagle (Soccer Simulation League 2D)                                        *
 * BASE SOURCE CODE RELEASE 2009                                                    *
 * Copyright (C) 1998-2009 WrightEagle 2D Soccer Simulation Team,                   *
 *                         Multi-Agent Systems Lab.,                                *
 *                         School of Computer Science and Technology,               *
 *                         University of Science and Technology of China, China.    *
 *                                                                                  *
 * This program is free software; you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License                      *
 * as published by the Free Software Foundation; either version 2                   *
 * of the License, or (at your option) any later version.                           *
 *                                                                                  *
 * This program is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    *
 * GNU General Public License for more details.                                     *
 *                                                                                  *
 * You should have received a copy of the GNU General Public License                *
 * along with this program; if not, write to the Free Software                      *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.  *
 ************************************************************************************/

#include "NoiseModel.h"
#include "ServerParam.h"
#include <fstream>
#include <cstdlib>

NoiseModel &NoiseModel::instance()
{
	static NoiseModel noise_model;
	return noise_model;
}

NoiseModel::NoiseModel()
{
	mMinBallSpeed = 0.01;
	mMaxBallSpeed = ServerParam::instance().ballSpeedMax();

	for (int i = 0; i < VEL_SIZE; i++)///ʼ5ű
		for (int j = 0; j < CYCLE_SIZE; j++)
		{
			mX[i][j] = 0.0;
			mXs[i][j] = 0.0;
			mY[i][j] = 0.0;
			mYs[i][j] = 0.0;
			mPxy[i][j] = 0.0;
		}

	ReadFile("data/noise_x", 1);
	ReadFile("data/noise_xs", 2);
	ReadFile("data/noise_y", 3);
	ReadFile("data/noise_ys", 4);
	ReadFile("data/noise_pxy", 5);
}


/**
 * ȡļݣ浽5
 *
 * @param file_name ļ
 * @param index ֱ5ļx, xs, y, ys, pxy
 * @return trueʾɹfalseʾʧ
 **
 * Read files and store them to 5 tables.
 * \param file_name.
 * \param index represents 5 files (x, xs, y, ys, pxy)
 * \return true if success, false otherwise.
 */
bool NoiseModel::ReadFile(const char* file_name, int index)
{
	std::ifstream in_file;
	in_file.open(file_name);
	if (!in_file)
	{
		PRINT_ERROR("open file error: " << file_name);
		exit(1);
	}

	int i = 0, j = 0;
	double tmp;
	while (in_file >> tmp)
	{
		if (j == 0 && tmp != 0.0)
		{
			PRINT_ERROR("read file error!");
			return false;
		}

		switch (index)
		{
		case 1:
			mX[i][j] = tmp;
			break;
		case 2:
			mXs[i][j] = tmp;
			break;
		case 3:
			mY[i][j] = tmp;
			break;
		case 4:
			mYs[i][j] = tmp;
			break;
		case 5:
			mPxy[i][j] = tmp;
			break;
		default:
			return false;
		}
		j++;
		if (j >= CYCLE_SIZE)
		{
			j = 0;
			i++;
		}
		if (i >= VEL_SIZE)
			break;
	}

	in_file.close();
	return true;
}

NoiseModel::~NoiseModel() {
}


/**
 * ٴСõӦ±
 *
 * @param speed ٶȵĴС
 * @return ±
 **
 * Get index from ball speed.
 * \param speed ball speed.
 * \return index.
 */
int NoiseModel::GetIndexFromSpeed(double speed)
{
	NormalizeSpeed(speed);

	int i;
	if (speed * 100 - (int)(speed * 100) >= 0.5)///뷨
		i = (int)(speed * 100);
	else
		i = (int)(speed * 100) - 1;
	return i;
}


/**
 * ٴС;õܶں󵽴õ
 *
 * @param speed ٶȵĴС
 * @param dis Ŀʼλõľ루ٷ
 * @return 
 **
 * Get cycles in need to a certain point most likely according to speed and distance.
 * \param speed ball speed.
 * \param dis distance between target and origin on the direction of ball velocity.
 * \return cycles in need most likely.
 */
int NoiseModel::GetBestCycle(double speed, double dis)
{
	int i, low, high, mid;
	i = GetIndexFromSpeed(speed);
	low = 0;
	high = 49;
	while (high - low > 1)
	{
		mid = (low + high) / 2;
		if (dis == mX[i][mid])
			return mid;
		else if (dis < mX[i][mid])
			high = mid;
		else
			low = mid;
	}
	if (fabs(dis - mX[i][low]) < fabs(dis - mX[i][high]))
		return low;
	else
		return high;
}


/**
 * ٶȣʱʹ
 *
 * @param ball_vel ٶ
 * @param max_rand ֵ
 * @param vel[] ²ٶȴŵ
 **
 * Generate more velocities, in use when considering random errors.
 * \param ball_vel initial velocity.
 * \param max_rand maximum random error.
 * \param vel will be filled with new generated velocities.
 */
void NoiseModel::GetMoreVelocity(Vector ball_vel, double max_rand, Vector vel[])
{
	///Ϊ8ȵĿٶֵ
	max_rand *= 0.5;
	double factor = 1.0 / SQRT_2;
	Vector r_0 (max_rand, 0.0);
	Vector r_90 (0.0, max_rand);
	Vector r_45 (max_rand * factor, max_rand * factor);
	Vector r_135 ( -max_rand * factor, max_rand * factor);

	vel[0] = ball_vel - r_0;
	vel[4] = ball_vel + r_0;
	vel[1] = ball_vel - r_45;
	vel[5] = ball_vel + r_45;
	vel[2] = ball_vel - r_90;
	vel[6] = ball_vel + r_90;
	vel[3] = ball_vel - r_135;
	vel[7] = ball_vel + r_135;
}


/**
 * õĳһĸܶ
 *
 * @param vel_index ٶȵ±꣬0.013.0
 * @param cyc_index ڵ±꣬049
 * @param pos õ
 * @return õĸܶ
 **
 * Get probability density on a certain point.
 * \param vel_index index which represents speed, from 0.01 to 3.0.
 * \param cyc_index index which represents cycle, from 0 to 49.
 * \param pos the point.
 * \return probability density on pos.
 */
double NoiseModel::GetPDF(int vel_index, int cyc_index, Vector pos)
{
	NormalizeIndex(vel_index, cyc_index);

	double x, y, x0, xs, y0, ys, pxy;
	x   = pos.X();
	y   = pos.Y();
	x0  = mX[vel_index][cyc_index];///xľֵ
	xs  = mXs[vel_index][cyc_index];///xķ
	y0  = mY[vel_index][cyc_index];///yľֵ
	ys  = mYs[vel_index][cyc_index];///yķ
	pxy = mPxy[vel_index][cyc_index];///ϵ

	double exp;
	exp = (x - x0) * (x - x0) / xs;
	exp -= 2 * pxy * (x - x0) * (y - y0) / sqrt(xs * ys);
	exp += (y - y0) * (y - y0) / ys;
	exp /= (-2 * (1 - pxy * pxy));
	exp = pow(M_E, exp);

	double pdf = exp / (2 * M_PI * sqrt(xs * ys) * (sqrt(1 - pxy * pxy)));

	if (pdf >= 0.0)
		return pdf;
	else
		return 0.0;
}


/**
 * ±õĳһΪԲĵԲĸʣػʱԲھȡ17
 *
 * @param vel_index ٶȵ±꣬0.013.0
 * @param cyc_index ڵ±꣬049
 * @param pos0 Բĵ
 * @param r Բİ뾶
 * @return 
 **
 * Get probability on a circle centered at a certain point. Checked 17 points when calculating double
 * integration.
 * \param vel_index index which represents speed, from 0.01 to 3.0.
 * \param cyc_index index which represents cycle, from 0 to 49.
 * \param pos0 center of the circle.
 * \param r radius of the circle.
 * \return probability.
 */
double NoiseModel::GetPossibilityFromIndex(int vel_index, int cyc_index, Vector pos0, double r)
{
	Vector pos[17];
	pos[0] = pos0;

	double buffer = -r;
	double factor = 1.0 / SQRT_2;
	for (unsigned i = 1; i <= 4; i++)///16㸳ֵ
	{
		pos[i]      = Vector(pos0.X() + buffer, pos0.Y());
		pos[i + 4]  = Vector(pos0.X(), pos0.Y() + buffer);
		pos[i + 8]  = Vector(pos0.X() + buffer * factor, pos0.Y() + buffer * factor);
		pos[i + 12] = Vector(pos0.X() + buffer * factor, pos0.Y() - buffer * factor);

		buffer += 0.5 * r;
		if (fabs(buffer) < FLOAT_EPS)
			buffer += 0.5 * r;
	}

	///ػ֣Ӷõ
	double ave = 0.0;
	for (unsigned i = 0; i < 17; i++)
		ave += GetPDF(vel_index, cyc_index, pos[i]);
	ave /= 17;
	double possibility = M_PI * r * r * ave;

	if (possibility >= 0.0 && possibility <= 1)
		return possibility;
	else if (possibility < 0.0)
		return 0.0;
	else
		return 1.0;
}


/**
 * ɳٶȺ;ڵõĳһΪԲĵԲĸ
 *
 * @param speed ĳٶ
 * @param cycle ʼк󾭹
 * @param goal_pos Բĵ
 * @param r Բİ뾶
 * @return 
 **
 * Get probability on a circle centered at a certain point.
 * \param speed ball speed.
 * \param cycle.
 * \param goal_pos center of the circle.
 * \param r radius of the circle.
 * \return probability.
 */
double NoiseModel::GetPossibilityFromSpeed(double speed, int cycle, Vector goal_pos, double r)
{
	int i = GetIndexFromSpeed(speed);
	return GetPossibilityFromIndex(i, cycle, goal_pos, r);
}


/**
 * õĳһΪԲĵԲĸ
 *
 * @param ball_vel ĳٶȣԣ
 * @param max_rand ܲ漴
 * @param cycle ʼк󾭹
 * @param pos_2_bvall Բʼλõ꣨goal_pos - ball_pos
 * @param r Բİ뾶
 * @return 
 **
 * Get probability on a circle centered at a certain point.
 * \param ball_vel ball velocity (in field coordinate system).
 * \param max_rand maximum random error.
 * \param cycle.
 * \param pos_2_ball relative position of center of the circle to ball.
 * \param r radius of the circle.
 * \return probability.
 */
double NoiseModel::GetPossibility(Vector ball_vel, double max_rand, int cycle, Vector pos_2_ball, double r)
{
	///TIMETEST("GetPossibility");
	double speed;
	Vector pos;

	if (max_rand == 0.0)///
	{
		speed = ball_vel.Mod();
		pos = pos_2_ball.Rotate(-ball_vel.Dir());
		return GetPossibilityFromSpeed(speed, cycle, pos, r);
	}

	Vector vel[8];
	GetMoreVelocity(ball_vel, max_rand, vel);

	///ÿٶȵĸʣȻȡƽ
	double ave = 0.0;
	for (unsigned i = 0; i < 8; i++)
	{
		speed = vel[i].Mod();
		pos = pos_2_ball.Rotate(-vel[i].Dir());
		ave += GetPossibilityFromSpeed(speed, cycle, pos, r);
	}
	ave /= 8;
	return ave;
}


/**
 * õxyıԵܶ
 *
 * @param vel_index ٶȵ±꣬0.013.0
 * @param cyc_index ڵ±꣬049
 * @param param xֵyֵ
 * @param is_x trueʾx
 * @return Եܶ
 **
 * Get marginal probability density for x or y.
 * \param vel_index index which represents speed, from 0.01 to 3.0.
 * \param cyc_index index which represents cycle, from 0 to 49.
 * \param param value of x or y.
 * \param is_x true represents x, and false represents y.
 * \return marginal probability density.
 */
double NoiseModel::GetPDFXY(int vel_index, int cyc_index, double param, bool is_x)
{
	NormalizeIndex(vel_index, cyc_index);

	double param0, params;
	if (is_x)
	{
		param0  = mX[vel_index][cyc_index];///xľֵ
		params  = mXs[vel_index][cyc_index];///xķ
	}
	else
	{
		param0  = mY[vel_index][cyc_index];///yľֵ
		params  = mYs[vel_index][cyc_index];///yķ
	}

	double exp, pdf;
	exp = -(param - param0) * (param - param0) / params / 2.0;
	exp = pow(M_E, exp);
	pdf = exp / sqrt(2.0 * M_PI * params);

	if (pdf >= 0.0)
		return pdf;
	else
		return 0.0;
}


/**
 * õĳһĸʣnewton-cotesʽ
 *
 * @param vel_index ٶȵ±꣬0.013.0
 * @param cyc_index ڵ±꣬049
 * @param low 
 * @param high 
 * @param is_x trueʾx
 * @return 
 **
 * Get probability on (low, high) through newton-cotes formula.
 * \param vel_index index which represents speed, from 0.01 to 3.0.
 * \param cyc_index index which represents cycle, from 0 to 49.
 * \param low
 * \param high
 * \param is_x true represents x axis, and false represents y axis.
 * \return probability.
 */
double NoiseModel::GetPXYFromIndex(int vel_index, int cyc_index, double low, double high, bool is_x)
{
	double h = (high - low) / 4.0;
	double possibility = 0.0;

	possibility += 7.0 * GetPDFXY(vel_index, cyc_index, low, is_x);
	possibility += 32.0 * GetPDFXY(vel_index, cyc_index, low + h, is_x);
	possibility += 12.0 * GetPDFXY(vel_index, cyc_index, low + 2.0 * h, is_x);
	possibility += 32.0 * GetPDFXY(vel_index, cyc_index, low + 3.0 * h, is_x);
	possibility += 7.0 * GetPDFXY(vel_index, cyc_index, high, is_x);
	possibility *= 2.0 * h / 45.0;
	return possibility;
}


/**
 * õĳһĸ
 *
 * @param vel_index ٶȵ±꣬0.013.0
 * @param cyc_index ڵ±꣬049
 * @param high 
 * @param is_x trueʾx
 * @return 
 **
 * Get probability on (-, high)
 * \param vel_index index which represents speed, from 0.01 to 3.0.
 * \param cyc_index index which represents cycle, from 0 to 49.
 * \param high
 * \param is_x true represents x axis, and false represents y axis.
 * \return probability.
 */
double NoiseModel::GetPXYFromIndex(int vel_index, int cyc_index, double high, bool is_x)
{
	NormalizeIndex(vel_index, cyc_index);

	double param0;
	if (is_x)
		param0 = mX[vel_index][cyc_index];
	else
		param0 = mY[vel_index][cyc_index];

	if (high == param0)
		return 0.5;
	else if(high < param0)
		return MinMax(0.0, 0.5 - GetPXYFromIndex(vel_index, cyc_index, high, param0, is_x), 1.0);
	else
		return MinMax(0.0, 0.5 + GetPXYFromIndex(vel_index, cyc_index, param0, high, is_x), 1.0);
}


/**
 * ٶֵõĳһĸ
 *
 * @param speed ٶ
 * @param cycle 
 * @param high 
 * @param is_x trueʾx
 * @return 
 *
 * Get probability on (-, high) via speed.
 * \param speed.
 * \param cycle.
 * \param high
 * \param is_x true represents x axis, and false represents y axis.
 * \return probability.
 */
double NoiseModel::GetPXYFromSpeed(double speed, int cycle, double high, bool is_x)
{
	int i = GetIndexFromSpeed(speed);
	return GetPXYFromIndex(i, cycle, high, is_x);
}


/**
 * õĳһxyֵĸ
 *
 * @param ball_vel ĳٶȣԣ
 * @param max_rand kick
 * @param cycle 
 * @param pos_2_ball Ŀλ
 * @param is_x trueʾx
 * @return 
 **
 * Get probability on interval from - to a certain point.
 * \param ball_vel ball velocity (in field coordinate system).
 * \param max_rand maximum random error.
 * \param cycle.
 * \param pos_2_ball relative position of center of the circle to ball.
 * \param is_x true represents x axis, and false represents y axis.
 * \return probability.
 */
double NoiseModel::GetPossibilityXY(Vector ball_vel, double max_rand, int cycle, Vector pos_2_ball, bool is_x)
{
	///TIMETEST("GetPossibilityXY");
	double speed, high;
	Vector pos;

	if (max_rand == 0.0)///
	{
		speed = ball_vel.Mod();
		pos = pos_2_ball.Rotate(-ball_vel.Dir());
		if (is_x)
			high = pos.X();
		else
			high = pos.Y();
		return GetPXYFromSpeed(speed, cycle, high, is_x);
	}

	Vector vel[8];
	GetMoreVelocity(ball_vel, max_rand, vel);

	///ÿٶȵĸʣȻȡƽ
	double ave = 0.0;
	for (unsigned i = 0; i < 8; i++)
	{
		speed = vel[i].Mod();
		pos = pos_2_ball.Rotate(-vel[i].Dir());
		if (is_x)
			high = pos.X();
		else
			high = pos.Y();
		ave += GetPXYFromSpeed(speed, cycle, high, is_x);
	}
	ave /= 8;
	return ave;
}


/**
 * ٶںϷٶȷΧڣ0.01 - 3.0
 *
 * @param speed ٶֵ
 *
 * Normalize give speed to legal range.
 * \param speed given speed and will be set to legal speed.
 */
void NoiseModel::NormalizeSpeed(double& speed)
{
	if (speed > mMaxBallSpeed)
		speed = mMaxBallSpeed;
	if (speed < mMinBallSpeed)
		speed = mMinBallSpeed;
}


/**
 * ֵںϷΧ
 *
 * @param vel_index ٶ
 * @param cyc_index 
 *
 * Normalize indexes to legal range.
 * \param vel_index index of velocity and will be set to legal index.
 * \param cyc_index index of cycle and will be set to legal index.
 */
void NoiseModel::NormalizeIndex(int& vel_index, int& cyc_index)
{
	if (vel_index < 0)
		vel_index = 0;
	if (vel_index > VEL_SIZE - 1)
		vel_index = VEL_SIZE - 1;

	if (cyc_index < 0)
		cyc_index = 0;
	if (cyc_index > CYCLE_SIZE - 1)
		cyc_index = CYCLE_SIZE - 1;
}
