#ifndef __Strategy_H__
#define __Strategy_H__

#include <vector>
#include "Types.h"
#include "Geometry.h"
#include "InterceptInfo.h"
#include "BehaviorBase.h"
#include "DecisionData.h"

class Agent;
class PlayerState;

/**
*  Strategy ϢʱҪõԼú룬Բܷ InfoState 棬Ƶ Agent 
*/
class Strategy: public DecisionData
{
public:
	Strategy(Agent & agent);
	virtual ~Strategy();

	void UpdateRoutine();

private:
	/**
	 * νصĽ
	 * fix interception points
	 */
	void DashGetBallFix();
	/**
	 * ʼ
	 * The initiation of the strategy
	 */
	void StrategyAnalyze();
	/**
	 * Ȩ
	 * the anvlyse of ball possession
	 */
	void BallPossessionAnalyse();
	bool SetPlayAnalyse();
	/**
	 * ԵǰƵķ
	 * the analyse of current situation
	 */
	void SituationAnalyse();

public:
	/**
	 * get current attack phase
	 * @return
	 */
    const AttackPhase & GetAttackPhase() const { return mAttackPhase; }
    /**
     * calculate current attack phase
     */
    AttackPhase CalcAttackPhase(const double & x) const;

private:
    double mOppForwardLine;
    double mOppMidLine;
    double mOppBackLine;
    AttackPhase mAttackPhase;

public:
    /**
     * õt_numŶѵĻλ
     * get the Situation Based Strategic Positioning for the teammate whose number is "t_num"
     */
	Vector GetTeammateSBSPPosition(int t_num,const Vector& ballpos,bool normal = false);

	/**
	 * tm_unumŶѵĿ귽
	 * get the aim angle of the teammate whose number is tm_unum
	 */
    AngleDeg CalcAimAngleForTeammate(const Unum &tm_unum);
    /**
     * жҵǷ㹻ȫdistanceԶʣһ
     * judge whether I can dash a "distance" with max power with a certain stamina left
     */
    bool IsMyStaminaEnough(double distance);

public:
    /**
     * жǷʵʿ
     * judge whether the ball is actural kickable
     * @return true when the ball is actural kickable
     */
	bool IsBallActuralKickable() const { return mIsBallActuralKickable;}
	/**
	 * жǷڡɡ״̬ǰ˿
	 * judge whether the ball is free
	 * @return true when the ball is free
	 */
	bool IsBallFree() const { return mIsBallFree; }
	/**
	 * ball_freeܳʱ
	 * get the cycles during the ball is free
	 * @return the cycles during the ball is free
	 */
	int  GetBallFreeCycleLeft() const { return mBallFreeCycleLeft; }
	/**
	 * жǷǱѿ
	 * judge whether the ball is kickable by a teammate
	 */
	bool IsTmKickable() const;

	/**
	 * жϵǰǷߵ
	 * judge whether I commit the penalty
	 */
	bool IsMyPenaltyTaken() const;

	/**
	 * ǰ˵ĺ룬+ʾԼ˿-ʾԷ0ʾû˿
	 * get the number of the person who controls the ball
	 * + means the ball is controlled by teammates
	 * - means the ball is controlled by opponents
	 * 0 means nobody controls the ball now
	 * @return the number of the person who controls the ball
	 */
	Unum GetController() const { return mController;  }
	void SetController(Unum controller) { mController = controller; }
	/**
	 * жǷڿ
	 * judge whether I can control the ball
	 */
    bool IsMyControl() const;
    /**
     * жǷʯڿ
     * judge whether the ball is controled by a teammeate
     * @return true when the ball is controled by a teammate
     */
	bool IsTmControl() const { return mController > 0; }
    /**
     * жǷʯڿ
     * judge whether the ball is controled by a opponent
     * @return true when the ball is controled by a opponent
     */
	bool IsOppControl() const { return mController < 0; }
	/**
	 * жϵǰǷ˿
	 * judge whther the ball is controled by nobody
	 * @return true when the ball is controled by nobody
	 */
	bool IsBallOutOfControl() const { return mController == 0; }
    /**
     * жǷϴο
     * judge whether the ball is controled by me at last time
     */
    bool IsLastMyControl() const;
    /**
     * жϴοԱǷǶ
     * judge whether the ball is controled by a teamate at last time
     * @return true when the ball is controled by me at last time
     */
	bool IsLastTmControl() const { return mLastController > 0; }
    /**
     * жϴοԱǷǶ
     * judge whether the ball is controled by a opponent at last time
     * @return true when the ball is controled by me at last time
     */
	bool IsLastOppControl() const { return mLastController < 0; }
	/**
	 * жϸ׶Ƿfree״̬˿
	 * judge whether the ball is free at last time
	 * @return true when the ball is free at last time
	 */
    bool IsLastBallFree() const { return mIsLastBallFree; }
    /**
     * ϴο˵ĺ
     * get the number of the person who controled the ball at last time
     * @return the number of the person who controled the ball at last time
     */
    Unum GetLastController() const { return mLastController; }
    /**
     * һζֿʱֵĺ
     * get the number of the opponent at last time
     * @return the number of the opponent at last time
     */
    Unum GetLastChallenger() const { return mLastChallenger; }
    /**
     * һfree״̬ʱ
     * get the latest time when the ball is free
     * @return the time when the ball is free
     */
    const Time & GetLastBallFreeTime() const { return mLastBallFreeTime; }

    /**
     * ǰĹ
     * get the current situation of Offensive and Defense
     * @return the current situation
     */
    const Situation & GetSituation() const { return mSituation; }
	void	SetSituation(Situation S){mSituation = S;}

	/**
	 * õ
	 * get the cycles during attacking in the penalty area
	 * @return the cycles during attacking in the penalty area
	 */
    int GetPenaltyAttackCycle() const { return mPenaltyAttackCycle; }
    /**
     * жǷӦԶ
     * judge whether we should employ arge persons against small
     * @return true when we employ arge persons against small
     */
    bool IsAttackInAdvantage() const { return mIsAttackInAdvantage; }

    /**
     * ȽסĵطԼ˺ͶԷȥص
     * get the first position where the ball can be interceptted
     * @return the position where the ball can be interceptted
     */
	const Vector & GetBallInterPos() const { return mBallInterPos; }

	/**
	 * 
	 * get the cycles after which the ball will be out
	 * @return the cycles after which the ball will be out
	 */
	int GetBallOutCycle() const { return mBallOutCycle; }
	/**
	 * ҷڷ״̬
	 * the cycles during our defending
	 * @return the cycles during our defending
	 */
    int GetDefenseCycle() const { return mDefenseCycle; }
    /**
     * ǰ߱block -- Ҫݵǰʱ״̬ϸڵϢ״̬㣬ֱӴӵǰ״̬
     *
     * @return
     */
	double GetControllerBlockedCycle() const { return mControllerBlockedCycle; }

	/**
	 * cycle_delayĵǰԱСڣΪcycle_delayʱѾ
	 * get my least interception cycles with cycle delay
	 * @return my least interception cycles with cycle delay
	 */
	int GetMyInterCycle() const { return mMyInterCycle; }
	void SetMyInterCycle(int cycle) { mMyInterCycle = cycle; }

	/**
	 * ǰԱȥѽ
	 * get the best position where I go to intercept the ball
	 */
	Vector GetMyInterPos();

	/**
	 * cycle_delayĶֵСڣΪcycle_delayʱѾ
	 * get opponents' least interception cycles with cycle delay
	 * @return opponents' least interception cycles with cycle delay
	 */
	int GetMinOppInterCycle() const { return mMinOppInterCycle; }
	/**
	 * cycle_delayĶѵСڣΪcycle_delayʱѾ
	 * get teammates' least interception cycles with cycle delay
	 * @return teammates' least interception cycles with cycle delay
	 */
	int GetMinTmInterCycle() const { return mMinTmInterCycle; }
	/**
	 * cycle_delayĶȵõ
	 * get the number of the teammate who can control the ball first with cycle delay
	 * @return the number of the teammate who can control the ball first
	 */
	Unum GetFastestTm() const { return mFastestTm; }
	/**
	 * cycle_delayĶȵõ
	 * get the number of the opponent who can control the ball first
	 * @return the number of the opponent who can control the ball first
	 */
	Unum GetFastestOpp() const { return mFastestOpp; }

	/**
	 * һȵõĶ
	 * Get the fastest teammate to a virtual ball.
	 */
	PlayerInterceptInfo GetFastestTmToVirtualBall(Vector ballpos, Vector ballvel, Unum & teammate, bool include_self = false,bool include_goalie = false) const;
	/**
	 * һȵõĶ
	 * Get the fastest opponent to a virtual ball.
	 */
	PlayerInterceptInfo GetFastestOppToVirtualBall(Vector ballpos, Vector ballvel, Unum & opponent) const;

	/**
	 * ۺϿпܵĶֵĽ
	 * get the opponents' interception cycles based on an overall consideration of various factors
	 * @return the opponents' interception cycles
	 */
	int GetSureOppInterCycle() const { return mSureOppInterCycle; }
	/**
	 * ۺϿпܵĶѵĽ
	 * get the teammates' interception cycles based on an overall consideration of various factors
	 * @return the teammates' interception cycles
	 */
	int GetSureTmInterCycle() const { return mSureTmInterCycle; }
	/**
	 * ۺϿпܵĽ
	 * get the interception cycles based on an overall consideration of various factors
	 * @return the interception cycles
	 */
	int GetSureInterCycle() const {return mSureInterCycle;}
	/**
	 * cycle_delayĶȵõ
	 * get the number of the teammate who can control the ball first without cycle delay
	 * @return
	 */
	Unum GetSureTm() const { return mSureTm; }
	/**
	 * cycle_delayĶȵõ
	 * get the number of the opponent who can control the ball first without cycle delay
	 * @return
	 */
	Unum GetSureOpp() const { return mSureOpp; }

	/**
	 * õָѵĿ귽
	 * get the aim angle of the teammate whose number is "unum"
	 * @param unum
	 * @return the aim angle
	 */
	const double & GetTeammateAimAngle(Unum unum) const
	{
		Assert(unum >= 1 && unum <= TEAMSIZE); return mTeammateAimAngle[unum];
	}

	void SetTeammateAimAngle(Unum unum, double angle)
	{
		Assert(unum >= 1 && unum <= TEAMSIZE); mTeammateAimAngle[unum] = angle;
	}

	/**
	 * õֵָĿ귽
	 * get the aim angle of the opponent whose number is "unum"
	 * @param unum
	 * @return the aim angle
	 */
	double GetOpponentAimAngle(Unum unum) const
	{
		Assert(unum >= 1 && unum <= TEAMSIZE); return mOpponentAimAngle[unum];
	}

	/**
	 * õָԱĿ귽
	 * get the aim angle of the player whose number is "unum"
	 * @param unum
	 * @return the aim angle
	 */
	double GetPlayerAimAngle(Unum unum) const
	{
		return (unum > 0) ? GetTeammateAimAngle(unum) : GetOpponentAimAngle(-unum);
	}

	/**
	 * õǰŵ÷ֵĶб
	 * get a list of teammates who may score a goal
	 */
	const std::vector<Unum>& GetFocusScoreTeammate();
	/**
	* õǰŵ÷ֵĶб
	* get a list of opponents who may score a goal
	*/
	const std::vector<Unum>& GetFocusScoreOpponent();

private:
	AngleDeg mTeammateAimAngle[TEAMSIZE + 1];
	AngleDeg mOpponentAimAngle[TEAMSIZE + 1];

	bool mIsBallActuralKickable;//Ƿʵʿ
	bool mIsBallFree;
	Unum mController; //+ʾԼ˿-ʾԷ0ʾû˿˿freeһ£freeָǰ˿򣬵˲һҪ򣬿صˣ
	Unum mChallenger; //Էߵ +

    bool mIsLastBallFree;
    Time mLastBallFreeTime;
    Unum mLastController; //¼ϴοˣϸڣڸ㣩Ҳϸ׶εģnewinfostate
    Unum mLastChallenger;

	///ǿcycle_delayСڣΪcycle_delayʱѾ
	int mMyInterCycle;
	int mMinTmInterCycle; /*öѲԼ*/
	int mMinOppInterCycle;
	int mMinIntercCycle;
	Unum mFastestTm;
	Unum mFastestOpp;

	//ۺϿпܵĽ
	int mSureTmInterCycle;
	int mSureOppInterCycle;
	int mSureInterCycle;
	Unum mSureTm;
	Unum mSureOpp;

	Vector mBallInterPos; //ȽסĵطԼ˺ͶԷȥص
	int mBallFreeCycleLeft; //ball_freeܳ
	int mBallOutCycle; //

	Situation mSituation;	//ǰĹ

    int mPenaltyAttackCycle; // 
    bool mIsAttackInAdvantage; // жǷΪԶ

	std::vector<Unum> mFocusScoreTm; //ܵ÷ֵĽ
	Time mFocusScoreTmTime; //һμ
	std::vector<Unum> mFocusScoreOpp; //ܵ÷ֵĽ
	Time mFocusScoreOppTime; //һμ

    int mDefenseCycle; // ҷڷ״̬
	double mControllerBlockedCycle; //ǰ߱block -- Ҫݵǰʱ״̬ϸڵϢ״̬㣬ֱӴӵǰ״̬

/** state analyseõϢ*/
public:
	/**
	 * ָԱĲɹ
	 * calculate the successful probability of the teammate whose number is "unum"
	 * @param unum
	 * @return the successful probability
	 */
    double ComputeWillTackleProb(const Unum & unum);
    /**
     * õĶ
     * get the number of the opponent who get closest to the ball
     * @return the number of the opponent
     */
    const Unum & GetClosestOpp2Ball() const { return mClosestOpp2Ball; }
    /**
     * õĶľ
     * get the distance between the ball and the opponent who get closest to the ball
     * @return the distance
     */
    const double & GetClosestOpp2BallDist() const { return mClosestOpp2BallDist; }

private:
    Unum mClosestOpp2Ball;
    double mClosestOpp2BallDist;

/** ڵĽӿں */
public:
	void SetPenaltyFirstStep(bool flag)    { mIsPenaltyFirstStep = flag; }
	bool IsPenaltyFirstStep() const        { return mIsPenaltyFirstStep; }
	/**
	 * õǰˣַʱΪ-1
	 * get the number of the player whose turn to a penalty kick
	 * it is -1 when the player is the opponent
	 * @return
	 */
	Unum GetPenaltyTaker() const    { return mPenaltyTaker; }
	/**
	 * ҷڼη
	 * determine which time it is for us to spot kick.
	 * @return
	 */
	int GetPenaltySetupTime() const { return mPenaltySetupTime; }

	/**
	 * жϵǰǷҷ
	 * judge whether it is turn for teammates to have a penalty kick
	 */
	bool IsOurPenaltyPlayMode() const;
	/**
	 * жϵǰǷҷ
	 * judge whether it is turn for opponents to have a penalty kick
	 */
	bool IsOppPenaltyPlayMode() const;
	/**
	 * жϵǰǷǵģʽ
	 * judge whether it is time for penalty now
	 */
	bool IsPenaltyPlayMode() const;

	/**
	 * take_timeҷ˭
	 * get the number of the teammate who should have a penalty kick at the "take_time" times
	 */
	int GetOurPenaltyTaker(int take_time);
	/**
	 * ԵǰӦ˭ķ
	 * analyze who should have a penalty kick
	 */
	void PenaltyAnalyze();

private:
	bool mIsPenaltyFirstStep; // setupʱõ
	Unum mPenaltyTaker; // ˣַʱΪ-1
	int mPenaltySetupTime; // ҷڼη
	int mPenaltyTakerSequence[TEAMSIZE + 1]; // ҷ˳ţֻһ
//==============================================================================
public:
	/** last behaviorĽӿ */
	/**
	 * ע: õĽӿڶõģbehavior*治ù
	 * */

	/**
	 * õڱactivebehaviorûΪ
	 * get the activebehavior saved last cycle
	 * @param type
	 * @return
	 */
	ActiveBehavior *GetLastActiveBehavior(BehaviorType type) const;

	/**
	 * õʵִеactivebehavior
	 * get the activebehavior in its actual execution last cycle
	 * @return
	 */
	ActiveBehavior *GetLastActiveBehaviorInAct() const { return mLastActiveBehavior[0]; }

	/**
	 * Ƿִtype͵activebehavior
	 * judge whether the type of the activebehavior performed last cycle is "type"
	 * @param type
	 * @return true when the type of the activebehavior performed last cycle is "type"
	 */
	bool IsLastActiveBehaviorInActOf(BehaviorType type) const {
		return GetLastActiveBehaviorInAct() && GetLastActiveBehaviorInAct()->GetType() == type;
	}

private:
	/**
	 * behavior*߳activebehavior -- planʱ
	 * save the best activebehavior when "plan" finishes.
	 * @param beh
	 */
	void SaveActiveBehavior(const ActiveBehavior & beh);

	friend class DecisionTree;

	void SaveActiveBehaviorList(const std::list<ActiveBehavior> & behavior_list);

	/**
	* ñʵִеactivebehavior -- excuteʱ
	* set the the type of the activebehavior in its actual execution this cycle when it is "excute"
	*/
	void SetActiveBehaviorInAct(BehaviorType type);

	void ResetSavedActiveBehavior();

private:
	ActiveBehavior *mActiveBehavior[BT_Max];
	ActiveBehavior *mLastActiveBehavior[BT_Max];

public:
    PlayerInterceptInfo mMyTackleInfo; //ԼĿ߷Χ󵽲ĽϢ
    /**
     * Լĳĳɴľ
     * calculate roughly the foward distance at some point towards a certain direction
     */
    double CalcForwardDistance(const Vector & pt, const double dribble_speed, AngleDeg angle, double fix, bool ourside, Unum *pOppUnum) const;
    /**
     * 㳯ĳԴԶ
     * fixcycdelayȣ󣬶ԷΧһԲCbpԲangleϵCеԲDDԲļԵĵ
     * estimate the forward distance with dribbling the ball at some point toward a certain direction
     */
    double CalcDribbleDistanceVsPlayer(const Vector & bp, const double dribble_speed, const PlayerState & player, AngleDeg angle, double fix) const;

private:
	class VirtualSelf: public PlayerState {
	public:
		VirtualSelf(const PlayerState & player): PlayerState(player) {}
		/**
		 * õ߷Χ
		 * get the farest distance to a player where he can tackle the ball
		 * @return
		 */
		const double & GetKickableArea() const { return ServerParam::instance().tackleDist(); }
	};
};

#endif
